/*
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
*/

#include <string>
#include <sstream>
#include <vector>
#include <iostream>
#include <stdlib.h>

#include "Object.hh"
#include "util.h"
#include "sgb.h"
#include "Matrice.hh"
#include "Sphere.hh"
#include "FileHandler.hh"
#include "XmlElement.hh"

Object::Object(FILE* parFp, int parEndian, int parLevel)
{
    head_t theHead;
    int i;
    
    myType      = NULL;
    myName      = NULL;
    myResource  = NULL;
    mySphere    = NULL;
    myLevel     = parLevel;
    myId        = -1;

    freadEndian(&theHead, sizeof(head_t), 1, parFp, parEndian);

    //FIXME perhaps this will work better if I use a uint32_t and just mask the values out of it
    if(!parEndian)
    {
        uint8_t tmp;
        tmp = theHead.subobjects;
        theHead.subobjects = theHead.matrix_number;
        theHead.matrix_number = tmp;

        tmp = theHead.unknown;
        theHead.unknown = theHead.matrices;
        theHead.matrices = tmp;
    }
    
    myMatrixNumber = theHead.matrix_number;
    myNumType       = theHead.num_of_type;
    
    //Get the type string
    if(theHead.str1)
        myType      = new std::string(getString(parFp, theHead.str1, NULL));
    if(theHead.str2)
        myName      = new std::string(getString(parFp, theHead.str2, NULL));
    if(theHead.str3)
        myResource  = new std::string(getString(parFp, theHead.str3, NULL));
    mySphere    = new Sphere(theHead.sphere);

    //An object never have a third string... or does it?
    if( theHead.str3 != 0 && myType->compare("OBJECT") != 0)
    {
        std::cout<<"Object of type: "<<*myType<<std::endl;
        std::cout<<"Object of type: "<<theHead.str1<<std::endl;
        throw;
    }
    
    //Object specific stuff
    if(myType->compare("OBJECT") == 0)
    {
        freadEndian(&myUserFlags, sizeof(uint32_t), 1, parFp, parEndian);

        //If this object is at level 1 we have a matrice, even though the data says 0....?
        if(parLevel == 1)
            theHead.matrices = 1;
    }

    //Read the given number of matrices
    for(i = 0; i < theHead.matrices; i++)
    {
        Matrice* theMatrice = new Matrice(parFp, parEndian, i);
        myMatrices.push_back(theMatrice);
    }

    //Object specific stuff
    if(myType->compare("LOD") == 0)
    {
        for(i = 0; i < theHead.subobjects; i++)
        {
            float theLodDistance;
            freadEndian(&theLodDistance, sizeof(float), 1, parFp, parEndian);
            myLodDistances.push_back(theLodDistance);
        }
    }
    
    //Try to read any object references
    for(i = 0; i < theHead.subobjects; i++)
    {
        uint32_t theReference;
        freadEndian(&theReference, sizeof(uint32_t), 1, parFp, parEndian);
        myChildNodes.push_back(followReference(parFp, theReference, parEndian, parLevel+1));
    }
}
        

Object::Object(XmlElement* parXml, int parLevel)
{
    myType      = NULL;
    myName      = NULL;
    myResource  = NULL;
    mySphere    = NULL;
    myLevel     = parLevel;

    if(parXml->getAttribute("type"))
    {
        myType = new std::string(parXml->getAttributeValue("type"));
    }
    
    if(parXml->getAttribute("Name"))
    {
        myName = new std::string(parXml->getAttributeValue("Name"));
    }

    XmlElement* theElem;
    int theNum;

    theElem = parXml->getChildNode("RESOURCE");
    if(theElem) {
        if(theElem->getAttribute("Filename"))
            myResource = new std::string(theElem->getAttributeValue("Filename"));
    }
    
    myMatrixNumber  = atoi(parXml->getAttributeValue("MatrixNumber", "-1").c_str());
    myNumType       = atoi(parXml->getAttributeValue("numType", "1").c_str());
    myUserFlags     = atoi(parXml->getAttributeValue("userflags", "0").c_str());
    
    //SPHERE
    theElem = parXml->getChildNode("SPHERE");
    if(theElem)
    {
        mySphere = new Sphere(theElem);
    }

    //Marix
    theNum = parXml->nChildNode("MATRIX");
    for(int i = 0; i < theNum; i++)
    {
        Matrice* theMatrice;
        theMatrice = new Matrice(parXml->getChildNode("MATRIX", i));
        myMatrices.push_back(theMatrice);
    }

    //CONTROL
    theElem = parXml->getChildNode("CONTROL");
    if(theElem)
    {
        std::string theDist = theElem->getAttributeValue("Distances", "0");
        size_t theStart = 0;
        size_t theEnd = 0;
        while(theStart != std::string::npos)
        {
            theEnd = theDist.find_first_of(' ', theStart);
            if(theEnd == std::string::npos)
                break;

            float theValue = atof(theDist.substr(theStart, theEnd-theStart).c_str());
            myLodDistances.push_back(theValue);

            theStart = theDist.find_first_not_of(' ', theEnd);
        }
    }

    //Subnodes
    theNum = parXml->nChildNode("NODE");
    for(int i = 0; i < theNum; i++)
        myChildNodes.push_back(new Object(parXml->getChildNode("NODE", i), parLevel+1));
}


Object::~Object()
{
    delete myType;
    delete myName;
    delete myResource;
    delete mySphere;
    
    //Matrices
    for(std::vector<Matrice*>::const_iterator cii = myMatrices.begin(); cii != myMatrices.end(); cii++)
    {
        delete (*cii);
    }
    
    //Child nodes
    std::vector<Object*>::const_iterator cij;
    for(cij = myChildNodes.begin(); cij != myChildNodes.end(); cij++)
    {
        delete (*cij);
    }
}

Object* Object::followReference(FILE* parFp, size_t ref, int parEndian, int parLevel)
{
    Object* theObject = NULL;
    size_t theStartPos = ftell(parFp);
    
    fseek(parFp, ref, SEEK_SET);
    theObject = new Object(parFp, parEndian, parLevel);
    fseek(parFp, theStartPos, SEEK_SET);
    return theObject;
}

std::string Object::toXml()
{
    std::stringstream   theStream;
    std::string         theRval;
    
    std::string theBaseIndent = "\t\t\t";
    if(myLevel == 2)
        theBaseIndent += "\t";
   
    theStream << theBaseIndent<<"<NODE type=\"" << *myType << "\" Name=\"" << *myName << "\"";
    theStream << " MatrixNumber=\"" << myMatrixNumber << "\"";
    theStream << " matrices=\"" << myMatrices.size() << "\"";
    theStream << " subobjects=\"" << myChildNodes.size() << "\"";
    theStream << " numType=\"" << myNumType << "\"";
    if(myType->compare("OBJECT") == 0)
        theStream << " userflags=\"" << myUserFlags << "\"";
    theStream << ">" << std::endl;
    if(myResource)
        theStream << theBaseIndent << "\t<RESOURCE Filename=\"" << *myResource << "\" />" << std::endl;

    //The Sphere
    if(mySphere)
        theStream << mySphere->toXml(myLevel);
    
    //Matrices
    for(std::vector<Matrice*>::const_iterator cii = myMatrices.begin(); cii != myMatrices.end(); cii++)
    {
        theStream << (*cii)->toXml(myLevel);
    }
    
    //Add Control Distanes if we have any
    if(myLodDistances.size() > 0)
    {
        theStream << theBaseIndent << "\t<CONTROL Distances=\"";
        for(std::vector<float>::const_iterator cii = myLodDistances.begin(); cii != myLodDistances.end(); cii++)
        {
            theStream << *cii<<" ";
        }
        theStream << "\" />" << std::endl;
    }

    //Child nodes
    std::vector<Object*>::const_iterator cij;
    for(cij = myChildNodes.begin(); cij != myChildNodes.end(); cij++)
    {
        theStream << (*cij)->toXml();
    }

    theStream << theBaseIndent<<"</NODE>" << std::endl;
    theRval = theStream.str();
    return theRval;
}

size_t Object::rawLength()
{
    size_t theSize = sizeof(head_t);
    if(myType->compare("OBJECT") == 0)
        theSize += sizeof(uint32_t);
    
    //Matrices
    theSize += sizeof(matrix_t)*myMatrices.size();
    
    //Distances
    if(myType->compare("LOD") == 0)
        theSize += sizeof(float)*myLodDistances.size();

    //References to sub objects
    theSize += sizeof(uint32_t)*myChildNodes.size();
    
    std::vector<Object*>::const_iterator cij;
    for(cij = myChildNodes.begin(); cij != myChildNodes.end(); cij++)
    {
        theSize += (*cij)->rawLength();
    }

    return theSize;
}

void Object::toRaw(FileHandler* parFh)
{
    head_t theHead;
    theHead.matrices        = myMatrices.size();
    theHead.matrix_number   = myMatrixNumber;
    theHead.num_of_type     = myNumType;
    theHead.sphere          = mySphere->sphere();
    theHead.str1            = 0;
    theHead.str2            = 0;
    theHead.str3            = 0;
    if(myType)
        theHead.str1 = parFh->addString(*myType);
    if(myName)
        theHead.str2 = parFh->addString(*myName);
    if(myResource)
        theHead.str3 = parFh->addString(*myResource);;
    theHead.subobjects  = myChildNodes.size();
    theHead.unknown     = 0x00;
    
    if(myType->compare("OBJECT") == 0 && myLevel == 1)
        theHead.matrices = 0;
    
    //FIXME perhaps this will work better if I use a uint32_t and just mask the values out of it
    if(!parFh->endian())
    {
        uint8_t tmp;
        tmp = theHead.subobjects;
        theHead.subobjects = theHead.matrix_number;
        theHead.matrix_number = tmp;

        tmp = theHead.unknown;
        theHead.unknown = theHead.matrices;
        theHead.matrices = tmp;
    }
    
    parFh->write(&theHead, sizeof(head_t), 1);
    
    
    //Userflags (if OBJECT)
    if(myType->compare("OBJECT") == 0)
        parFh->write(&myUserFlags, sizeof(uint32_t), 1);
    
    //Matrices
    for(std::vector<Matrice*>::const_iterator cii = myMatrices.begin(); cii != myMatrices.end(); cii++)
    {
        (*cii)->toRaw(parFh);
    }
    
    //Distances (if LOD)
    if(myType->compare("LOD") == 0)
    {
        for(std::vector<float>::const_iterator cii = myLodDistances.begin(); cii != myLodDistances.end(); cii++)
        {
            parFh->write(&(*cii), sizeof(float), 1);
        }
    }
    
    
    //References
    std::vector<Object*>::const_iterator cij;
    uint32_t theOffset = parFh->tell() + (myChildNodes.size()*sizeof(uint32_t));
    for(cij = myChildNodes.begin(); cij != myChildNodes.end(); cij++)
    {
        parFh->write(&theOffset, sizeof(uint32_t), 1);
        theOffset += (*cij)->rawLength();
    }

    //Sub Objects
    for(cij = myChildNodes.begin(); cij != myChildNodes.end(); cij++)
    {
        (*cij)->toRaw(parFh);
    }
}
