/*
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
*/

#include <string>
#include <sstream>
#include <iostream>
#include <map>
#include <stdlib.h>
#include <math.h>

#include "ObjectId.hh"
#include "util.h"
#include "sgb.h"
#include "FileHandler.hh"
#include "XmlElement.hh"
        
int ObjectId::hex2int(std::string parHex)
{
    int theRval = 0;
    if(parHex.length() < 3)
        return atoi(parHex.c_str());

    if (!((parHex.at(0) == '0') && ((parHex.at(1) == 'x') || (parHex.at(1) == 'X'))))
        return atoi(parHex.c_str());

    int theLength = parHex.length();
    for(int i = theLength-1; i > 1; i--)
    {
        int pval = 0;
        char c = parHex.at(i);
        if(c >= '0' && c <= '9')
            pval = c-'0';
        if(c >= 'A' && c <= 'F')
            pval = c-'A';
        if(c >= 'a' && c <= 'f')
            pval = c-'a';
        theRval += pow(16,((theLength-1)-i))*pval;
    }
    return theRval;
}

ObjectId::ObjectId(FILE* parFp, int parEndian)
{
    myStr1 = NULL;
    myStr2 = NULL;
    myStr3 = NULL;
    
    objhead_t   theObjHead;
    uint32_t    theBlockSize;
    size_t      theCurrentPosition = ftell(parFp);
    freadEndian(&theBlockSize, sizeof(uint32_t), 1, parFp, parEndian);
    freadEndian(&theObjHead, sizeof(objhead_t), 1, parFp, parEndian);
    
    myNumber    = theObjHead.no;
    myUnknown   = theObjHead.unknown;
    myNumType   = theObjHead.num_of_type;

    if(!parEndian)
        myUnknown = swapBytes(myUnknown);

    //Get the strings
    if(theObjHead.str1)
        myStr1 = new std::string(getString(parFp, theObjHead.str1, NULL));
    if(theObjHead.str2)
        myStr2 = new std::string(getString(parFp, theObjHead.str2, NULL));
    if(theObjHead.str3)
        myStr3 = new std::string(getString(parFp, theObjHead.str3, NULL));
    
    //There is always one object following
    myObject = new Object(parFp, parEndian);
    fseek(parFp, theCurrentPosition+theBlockSize, SEEK_SET);
    //printf("Position: %ld -> %ld (%ld)\n", theCurrentPosition, ftell(parFp), ftell(parFp)-theCurrentPosition);
}
        
ObjectId::ObjectId(XmlElement* parXml)
{
    if(!parXml)
        throw;

    myStr1 = NULL;
    myStr2 = NULL;
    myStr3 = NULL;
    myObject = NULL;

    myNumber = atoi(parXml->getAttributeValue("no").c_str());
    myNumType = atoi(parXml->getAttributeValue("numType").c_str());

    //FIXME read a text hex value
    std::string theUnknown = parXml->getAttributeValue("unknown", "0x0");
    myUnknown = hex2int(theUnknown);

    if(parXml->getAttribute("str1"))
    {
        myStr1 = new std::string(parXml->getAttributeValue("str1"));
    }
    if(parXml->getAttribute("str2"))
    {
        myStr2 = new std::string(parXml->getAttributeValue("str2"));
    }
    if(parXml->getAttribute("str3"))
    {
        myStr3 = new std::string(parXml->getAttributeValue("str3"));
    }

    myObject = new Object(parXml->getChildNode("NODE"));

}


ObjectId::~ObjectId()
{
    delete myStr1;
    delete myStr2;
    delete myStr3;
    delete myObject;
}


std::string ObjectId::toXml()
{
    std::stringstream   theStream;
    std::string         theRval;

    theStream << "\t\t<OBJ_ID no=\"" << myNumber << "\"";
    theStream << " numType=\"";
    theStream << myNumType;
    theStream << "\"";
    theStream << " unknown=\"";
    theStream << std::hex<<std::showbase;
    theStream << myUnknown;
    theStream << "\"";
    if(myStr1)
        theStream << " str1=\""<<*myStr1<<"\"";
    if(myStr2)
        theStream << " str2=\""<<*myStr2<<"\"";
    if(myStr3)
        theStream << " str3=\""<<*myStr3<<"\"";
    theStream << ">" << std::endl;
    if(myObject)
        theStream << myObject->toXml();
    theStream << "\t\t</OBJ_ID>" << std::endl;

    theRval = theStream.str();
    return theRval;
}


size_t ObjectId::rawLength()
{
    size_t theSize = sizeof(uint32_t)+sizeof(objhead_t)+myObject->rawLength();
    //printf("Block length: %ld\n", theSize);
    return theSize;
}

void ObjectId::toRaw(FileHandler* parFh)
{
    uint32_t    theBlockSize;
    objhead_t   theHead;
    
    theBlockSize        = rawLength();
    theHead.no          = myNumber;
    theHead.num_of_type = myNumType;
    theHead.str1        = 0;
    theHead.str2        = 0;
    theHead.str3        = 0;
    
    if(!myStr3)
    {
        if(myStr2)
            theHead.str2    = parFh->addString(*myStr2);
        if(myStr1)
            theHead.str1    = parFh->addString(*myStr1);
    }
    if(myStr1 && myStr2 && myStr3)
    {
        if(myStr2)
            theHead.str2    = parFh->addString(*myStr2);
        if(myStr1)
            theHead.str1    = parFh->addString(*myStr1);
        if(myStr3)
            theHead.str3    = parFh->addString(*myStr3);
    }
    else
    {
        if(myStr1)
            theHead.str1    = parFh->addString(*myStr1);
        if(myStr2)
            theHead.str2    = parFh->addString(*myStr2);
        if(myStr3)
            theHead.str3    = parFh->addString(*myStr3);
    }
    theHead.unknown = myUnknown;
    if(!parFh->endian())
        theHead.unknown = swapBytes(theHead.unknown);
    
    parFh->write(&theBlockSize, sizeof(uint32_t), 1);
    parFh->write(&theHead, sizeof(objhead_t), 1);
    myObject->toRaw(parFh);
}
