/*
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
*/

#include <string>
#include <sstream>
#include <vector>
#include <iostream>

#include "Partition.hh"
#include "sgb.h"
#include "util.h"
#include <stddef.h>
#include "FileHandler.hh"
#include "XmlElement.hh"

int Partition::getMaxId()
{
    int theId = myId;
    std::vector<Partition*>::const_iterator cii;
    for(cii = myChildPartitions.begin(); cii != myChildPartitions.end(); cii++)
    {
        int theChildMax = (*cii)->getMaxId();
        if(theChildMax > theId)
            theId = theChildMax;
    }
    return theId;
}

Partition::Partition(FileHandler* parFh, int parId)
{
    flathead_t theHead;

    parFh->read(&theHead, sizeof(flathead_t), 1);
    
    myFlag  = 0;
    myId    = parId;
    myBox   = new AxisAlignedBoundingBox(theHead.min, theHead.max);

    int32_t theLength = theHead.length;
    if(theLength < 0)
    {
        myFlag = 1;
        theLength = (theLength^0xFFFFFFFF)+1;
    }
    
    theLength = theHead.num;
    if (theLength < 0)
    {
        //myFlag = 1;
        theHead.num = (theHead.num ^ 0xFFFFFFFF) + 1;
    }


    if (theLength > 0 && theHead.length>0 && theHead.length < 1000000 && theHead.num > 0)
    {
        unsigned int i;
        for(i = 0; i < theHead.num; i++) {
            PartitionChild* p = new PartitionChild(parFh);
            myObjects.push_back(p);
        }
    }
    
    //If not at end read sub partitions
    int thePartitionId = getMaxId()+1;
    while(rawLength() < (size_t)theLength)
    {
        Partition* thePartition = new Partition(parFh, thePartitionId);
        myChildPartitions.push_back(thePartition);
        thePartitionId = getMaxId()+1;
    }
}
        
Partition::Partition(XmlElement* parXml)
{
    if(!parXml)
        throw;

    myId = atoi(parXml->getAttributeValue("no", "0").c_str());

    XmlElement* theElem;
    //AABOX
    myBox = new AxisAlignedBoundingBox(parXml->getChildNode("AABBOX"));

    //CHILD_PARTITIONS IDs=
    theElem = parXml->getChildNode("CHILD_PARTITIONS");
    if(theElem)
    {
        std::string theSubPartitions = theElem->getAttributeValue("IDs", "NONE");
        if(theSubPartitions.compare("NONE") != 0)
        {
            size_t theStart = 0;
            size_t theEnd = 0;
            while(theStart != std::string::npos)
            {
                theEnd = theSubPartitions.find_first_of(' ', theStart);
                if(theEnd == std::string::npos)
                    break;

                std::string theId = theSubPartitions.substr(theStart, theEnd-theStart);
                XmlElement* theP = parXml->parent()->findChildwithAttribute("PARTITION_ID", "no", theId);
                if(theP)
                    myChildPartitions.push_back(new Partition(theP));

                theStart = theSubPartitions.find_first_not_of(' ', theEnd);
            }
        }
    }

    //Child objects
    int theNum = parXml->nChildNode("CHILD_OBJ");
    for(int i = 0; i < theNum; i++)
        myObjects.push_back(new PartitionChild(parXml->getChildNode("CHILD_OBJ", i)));
}


Partition::~Partition()
{
    if(myBox)
        delete myBox;
    
    //Delete Child Objects
    std::vector<PartitionChild*>::const_iterator cii;
    for(cii = myObjects.begin(); cii != myObjects.end(); cii++)
    {
        delete (*cii);
    }

    //Delete Child Partitions
    std::vector<Partition*>::const_iterator cij;
    for(cij = myChildPartitions.begin(); cij != myChildPartitions.end(); cij++)
    {
        delete (*cij);
    }
}

std::string Partition::toXml()
{
    std::stringstream   theStream;
    std::stringstream   theChildStream;
    std::string         theRval;
    
    theStream << "\t\t<PARTITION_ID no=\"" << myId << "\">" << std::endl;
    if(myBox)
    {
        theStream << myBox->toXml()->toXml(true, 1);
    }
    
    std::vector<PartitionChild*>::const_iterator cii;
    for(cii = myObjects.begin(); cii != myObjects.end(); cii++)
    {
        theStream << (*cii)->toXml();
    }

    theStream << "\t\t\t<CHILD_PARTITIONS IDs=\"";
    std::vector<Partition*>::const_iterator cij;
    for(cij = myChildPartitions.begin(); cij != myChildPartitions.end(); cij++)
    {
        theStream << (*cij)->id()<<" ";
        theChildStream << (*cij)->toXml();
    }
    if(myChildPartitions.size() == 0)
        theStream << "NONE";
    theStream << "\" />" << std::endl;
    
    theStream << "\t\t</PARTITION_ID>" << std::endl;
    
    //Append child partitions afterwards
    theStream << theChildStream.str();
    
    theRval = theStream.str();
    return theRval;
}

int Partition::id()
{
    return myId;
}

size_t Partition::rawLength()
{
    size_t mySize = 32 + (myObjects.size()*64);
    std::vector<Partition*>::const_iterator cij;
    for(cij = myChildPartitions.begin(); cij != myChildPartitions.end(); cij++)
    {
        mySize += (*cij)->rawLength();
    }
    return mySize;
}

void Partition::toRaw(FileHandler* parFh)
{
    flathead_t theHead;
    theHead.max = myBox->max();
    theHead.min = myBox->min();
    theHead.num = myObjects.size();
    theHead.length = rawLength();
    
    if(myFlag)
        theHead.length = (theHead.length^0xFFFFFFFF)+1;
    parFh->write(&theHead, sizeof(flathead_t), 1);
    
    //All child objects
    std::vector<PartitionChild*>::const_iterator cii;
    for(cii = myObjects.begin(); cii != myObjects.end(); cii++)
    {
        (*cii)->toRaw(parFh);
    }
    
    //All child partitions
    std::vector<Partition*>::const_iterator cij;
    for(cij = myChildPartitions.begin(); cij != myChildPartitions.end(); cij++)
    {
        (*cij)->toRaw(parFh);
    }
}
