/*
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
*/

#include <iostream>
#include <fstream>
#include <string>

#include "SgbData.hh"
#include "FileHandler.hh"
#include "XmlElement.hh"
#include "util.h"


SgbData::SgbData():
        mySgbSection(),
        myOcclSection(),
        myNodeSection(),
        myFlatSection(),
        mySummSection(std::string("SUMM_SECTION"), SECTION_SUMM),
        myEndSection()
{
}


std::string SgbData::toXml()
{
    std::string         theRval;
    std::stringstream   theStream;
    XmlElement* theElement;

    theStream << "<SGB>"<<std::endl;
    //SGB
        theElement = mySgbSection.toXml();
        theStream << theElement->toXml(true, 1);
        delete theElement;
    //OCLL
        theElement = myOcclSection.toXml();
        theStream << theElement->toXml(true, 1);
        delete theElement;
    //NODE
    theStream << myNodeSection.toXml();
    //FLAT
    theStream << myFlatSection.toXml();
    //SUMM
    theStream << mySummSection.toXml();
    //END
    theStream << myEndSection.toXml();

    theStream << "</SGB>"<<std::endl;
    theRval = theStream.str();
    return theRval;
}



int SgbData::readSection(FILE* fp, int endian, block_t* theSection)
{
    fseek(fp, theSection->start, SEEK_SET);

    section_head theHead;
    if(theSection->name != SECTION_DATA) {
        //Lets verify that we are at the right position
        fread(&theHead, sizeof(section_head), 1, fp);
        if(endian) {
            theHead.name = swapBytes(theHead.name);
        }
        else if(theSection->name != SECTION_SGB) {
            theHead.length = swapBytes(theHead.length);
            theHead.blocks = swapBytes(theHead.blocks);
        }

        if(theSection->name != theHead.name)
        {
            printf("Wrong offset!!\n");
            return 1;
        }

        printf("%.4s\n", (char*)&theHead.name);
        printf("blocks: %d\n", theHead.blocks);
        printf("length: %d\n", theHead.length);
    }

    FileHandler theFh;
    theFh.fp(fp);
    theFh.endian(endian);

    if(theSection->name == SECTION_NODE)
    {
        myNodeSection.readSection(fp, endian, theHead.blocks);
    }
    if(theSection->name == SECTION_SUMM)
    {
        mySummSection.readSection(fp, endian, theHead.blocks);
    }
    else if(theSection->name == SECTION_OCCL) //OCCL
    {
        myOcclSection.readSection(fp, endian, theHead.blocks);
    }
    else if(theSection->name == SECTION_FLAT)
    {
        myFlatSection.readSection(&theFh, theSection);
    }
    return 0;
}

void SgbData::readRaw(std::string parFile)
{
    FILE* theFp;
    uint32_t data;
    int endian = 0;
    
    theFp = fopen(parFile.c_str(), "rb");

    fread(&data, sizeof(uint32_t), 1, theFp);
    if(data != SECTION_SGB) {
        if(swapBytes(data) != SECTION_SGB) {
            printf("Unknown file type\n");
            fclose(theFp);
            return;
        }
        endian = 1;
    }

    //Scan file for sections
    block_t* theSections = findSections(theFp, endian);

    block_t* b = theSections;
    while(b)
    {
        readSection(theFp, endian, b);
        b = b->next;
    }
    fclose(theFp);
    freeSections(&theSections);
}


void SgbData::saveRaw(std::string parFile, int parEndian)
{
    FILE* theFp;
    theFp = fopen(parFile.c_str(), "wb");
    FileHandler theFh;
    theFh.fp(theFp);
    theFh.endian(parEndian);
    
    //Get the total size so we can calculate data offsets
    size_t theDataLen = 0;
    theDataLen += mySgbSection.rawLength();
    theDataLen += myOcclSection.rawLength();
    theDataLen += myNodeSection.rawLength();
    theDataLen += myFlatSection.rawLength();
    theDataLen += mySummSection.rawLength();
    theDataLen += myEndSection.rawLength();
    theFh.length(theDataLen);
    
    //Write the file
    mySgbSection.toRaw(&theFh);
    myOcclSection.toRaw(&theFh);
    myNodeSection.toRaw(&theFh);
    myFlatSection.toRaw(&theFh);
    mySummSection.toRaw(&theFh);
    myEndSection.toRaw(&theFh);
    theFh.writeData();
    
    fclose(theFp);
}


void SgbData::readXml(std::string parFile)
{
    std::string buf;
    std::string line;
    std::ifstream in(parFile.c_str());
    while(std::getline(in,line))
        buf += line;
    XmlElement* root = XmlElement::parseString(buf);
    
    myOcclSection.readSection(root->getChildNode(std::string("OCCL_SECTION")));
    myNodeSection.readSection(root->getChildNode(std::string("SECTION_NODE")));
    myFlatSection.readSection(root->getChildNode(std::string("FLAT_SECTION")));
    mySummSection.readSection(root->getChildNode(std::string("SUMM_SECTION")));
    delete root;
}


void SgbData::saveXml(std::string parFile)
{
    FILE* theFp;
    std::string theData;
    theFp = fopen(parFile.c_str(), "wb");
    theData = toXml();
    fwrite(theData.c_str(), theData.length(), 1, theFp);
    fclose(theFp);
}

