/*
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
*/

#include <string>
#include <sstream>
#include <iostream>
#include <stdlib.h>

#include "XmlElement.hh"

XmlElement::XmlElement(std::string parName, std::string parValue, NodeType parType)
{
    myParent        = NULL;
    myFirstAttr     = NULL;
    myFirstChild    = NULL;
    myPrevSibling   = NULL;
    myNextSibling   = NULL;
    myType          = parType;

    myName  = parName;
    myValue = parValue;
}
        
XmlElement::~XmlElement()
{
    XmlElement* theElement;

    //Delete all attributes
    theElement = myFirstAttr;
    while(theElement && theElement->myNextSibling)
    {
        XmlElement* theSibling = theElement->myNextSibling;
        delete theElement;
        theElement = theSibling;
    }
    delete theElement;

    //Delete all child elements
    theElement = myFirstChild;
    while(theElement && theElement->myNextSibling)
    {
        XmlElement* theSibling = theElement->myNextSibling;
        delete theElement;
        theElement = theSibling;
    }
    delete theElement;
}
        
XmlElement::XmlElement(XmlTookenizer* parTokenizer, NodeType parType)
{
    myParent        = NULL;
    myFirstAttr     = NULL;
    myFirstChild    = NULL;
    myPrevSibling   = NULL;
    myNextSibling   = NULL;
    myType          = parType;

    XmlTookenizer::XmlTooken* theTooken = parTokenizer->getToken();
    myName  = theTooken->myData;
    myValue = std::string("");
  
    //Get attributes if any
    while(true)
    {
        theTooken = parTokenizer->getToken(theTooken);
        if(theTooken->myType == XmlTookenizer::TOKEN_TEXT)
        {
            std::string theName;
            std::string theValue;

            theName = theTooken->myData;
            theTooken = parTokenizer->getToken(theTooken);
            if(theTooken->myType != XmlTookenizer::TOKEN_EQUAL)
            {
                std::cerr << "Attribute name not followed by equal sign (" << theTooken->myType <<")" << std::endl;
                throw;
            }
            theTooken = parTokenizer->getToken(theTooken);
            if(!theTooken || !(theTooken->myType == XmlTookenizer::TOKEN_TEXT || theTooken->myType == XmlTookenizer::TOKEN_QUOTEDTEXT))
            {
                std::cerr << "No attribute value" << std::endl;
                throw;
            }
            theValue = theTooken->myData;

            addAttribute(theName, theValue);
        }
        else if(theTooken->myType == XmlTookenizer::TOKEN_TAGEND)
        {
            //End of the start tag
            break;
        }
        else if(theTooken->myType == XmlTookenizer::TOKEN_EMPTYELEM)
        {
            //Empty element closing, means no body
            delete theTooken;
            return;
        }
        else
        {
            std::cout << "What is this? ("<< theTooken->myType <<")\n";
            throw;
        }
    }
        
    //Trying to read the body

    //Do we have any child tags?
    theTooken = parTokenizer->getToken(theTooken);
    while(theTooken && theTooken->myType == XmlTookenizer::TOKEN_TAGSTART)
    {
        XmlElement *e;
        e = new XmlElement(parTokenizer);
        appendChild(e);
        theTooken = parTokenizer->getToken(theTooken);
    }
    
    if(theTooken && theTooken->myType == XmlTookenizer::TOKEN_TAGCLOSE)
    {
        //We are done here
        theTooken = parTokenizer->getToken(theTooken); //The name
        theTooken = parTokenizer->getToken(theTooken); //The end tag
        //TODO make sure the close tag is for this tag.... otherwise we have broken xml
        delete theTooken;
        return;
    }
    else
    {
        std::cerr << "Found something else? " << theTooken->myType << " : " << theTooken->myData << std::endl;
        throw;
    }
}
        
XmlElement* XmlElement::parseString(std::string parData)
{
    XmlElement *e;
    XmlTookenizer* theTookenizer = new XmlTookenizer(parData);

    XmlTookenizer::XmlTooken* theTooken = theTookenizer->getToken();
    if(theTooken->myType != XmlTookenizer::TOKEN_TAGSTART)
    {
        std::cerr << "Found something other than tag start...\n";
        throw;
    }
    delete theTooken;
    e = new XmlElement(theTookenizer);
    delete theTookenizer;
    return e;
}

std::string XmlElement::toXml(bool parIndent, int parLevel)
{
    std::string         theRval;
    std::stringstream   theStream;
    if(myType == TYPE_ATTRIBUTE)
    {
        theStream << " " << myName << "=\"" << myValue << "\"";
    }
    else if(myType == TYPE_ELEMENT)
    {
        if(parIndent)
            theStream << indent(parLevel);
        theStream << "<" << myName;

        XmlElement* theAttribute = myFirstAttr;
        while(theAttribute)
        {
            theStream << theAttribute->toXml(parIndent, parLevel+1);
            theAttribute = theAttribute->myNextSibling;
        }

        if(myFirstChild == NULL)
            theStream << " />";
        else
        {
            theStream << ">" << std::endl;

            XmlElement* theChild = myFirstChild;
            while(theChild)
            {
                theStream << theChild->toXml(parIndent, parLevel+1);
                theChild = theChild->myNextSibling;
            }

            if(parIndent)
                theStream << indent(parLevel);
            theStream << "</" << myName << ">";
        }
        theStream << std::endl;
    }

    theRval = theStream.str();
    return theRval;
}
        

void XmlElement::addAttribute(std::string parName, std::string parValue)
{
    XmlElement* theNewAttribute = new XmlElement(parName, parValue, TYPE_ATTRIBUTE);
    theNewAttribute->myParent = this;
    if(myFirstAttr == NULL)
    {
        myFirstAttr = theNewAttribute;
    }
    else
    {
        XmlElement* theAttribute = myFirstAttr;
        while(theAttribute->myNextSibling)
            theAttribute = theAttribute->myNextSibling;

        theAttribute->myNextSibling = theNewAttribute;
        theNewAttribute->myPrevSibling = theAttribute;
        theNewAttribute->myNextSibling = NULL;
    }
}


void XmlElement::appendChild(XmlElement* parChild)
{
    parChild->myParent = this;
    
    if(myFirstChild == NULL)
        myFirstChild = parChild;
    else
    {
        XmlElement* theNode = myFirstChild;
        while(theNode->myNextSibling)
            theNode = theNode->myNextSibling;

        theNode->myNextSibling = parChild;
        parChild->myPrevSibling = theNode;
        parChild->myNextSibling = NULL;
    }
}


XmlElement* XmlElement::getAttribute(std::string parName)
{
    XmlElement* theAttribute = myFirstAttr;
    while(theAttribute)
    {
        if(theAttribute->myName.compare(parName) == 0)
        {
            return theAttribute;
        }
        theAttribute = theAttribute->myNextSibling;
    }
    return NULL;
}


std::string XmlElement::getAttributeValue(std::string parName, std::string parDefault)
{
    XmlElement* theAttribute = getAttribute(parName);
    if(!theAttribute)
        return parDefault;
    return theAttribute->myValue;
}


XmlElement* XmlElement::getChildNode(std::string parName, int parNum)
{
    int thePos = 0;
    XmlElement* theChild = myFirstChild;
    while(theChild)
    {
        if(theChild->myName.compare(parName) == 0)
        {
            if(thePos == parNum)
                return theChild;
            thePos++;
        }
        theChild = theChild->myNextSibling;
    }
    return NULL;
}

unsigned int XmlElement::nChildNode(std::string parName)
{
    int thePos = 0;
    XmlElement* theChild = myFirstChild;
    while(theChild)
    {
        if(theChild->myName.compare(parName) == 0)
        {
            thePos++;
        }
        theChild = theChild->myNextSibling;
    }
    return thePos;
}


XmlElement* XmlElement::findChildwithAttribute(std::string parName, std::string parAttName, std::string parAttValue)
{
    XmlElement* theChild = myFirstChild;
    while(theChild)
    {
        if(theChild->myName.compare(parName) == 0)
        {
            if(theChild->getAttributeValue(parAttName).compare(parAttValue) == 0)
                return theChild;
        }
        theChild = theChild->myNextSibling;
    }
    return NULL;
}


std::string XmlElement::indent(int parLevel)
{
    static std::string theIndent;
    theIndent.resize(parLevel*2, ' ');
    return theIndent;
}


//XmlTookenizer
//---------------

int XmlElement::XmlTookenizer::findNonWhiteSpace()
{
    char c = myData.at(myPosition);
    while(c == '\n' || c == '\r' || c == '\t' || c == ' ')
    {
        myPosition++;
        c = myData.at(myPosition);
    }
    return myPosition;
}

int XmlElement::XmlTookenizer::findEndString()
{
    int thePosition = myPosition;
    char myQuote = 0;

    char c = myData.at(thePosition);
    if(c == '\'' || c == '"')
    {
        myQuote = c;
        thePosition++;
    }
    
    while(true)
    {
        c = myData.at(thePosition);
        if(myQuote)
        { 
            if(c == myQuote)
                break;
        }
        else if(c == ' ' || c == '\r' || c == '\n' || c == '\t' || c == '/' || c == '<' || c == '>' || c == '=')
        {
            break;
        }
        thePosition++;
    }

    return thePosition;
}

XmlElement::XmlTookenizer::XmlTookenizer(std::string parData)
{
    myData = parData;
    myPosition = 0;
}


XmlElement::XmlTookenizer::XmlTooken* XmlElement::XmlTookenizer::getToken(XmlTooken* parTooken)
{
    int startPos = findNonWhiteSpace();
    char c = myData.at(startPos);
    XmlTooken* myTooken = NULL;
    if(parTooken)
        myTooken = parTooken;
    else
        myTooken = new XmlTooken();
            
    int theStartPos;
    int theEndPos;
    switch(c)
    {
        case '<':
            //Peek forward
            if(myData.at(startPos+1) == '/')
            {
                //End
                myTooken->myType = TOKEN_TAGCLOSE;
                myPosition+=2;
            }
            else
            {
                //Start
                myTooken->myType = TOKEN_TAGSTART;
                myPosition++;
            }
            break;
        case '>':
            //End tag
            myTooken->myType = TOKEN_TAGEND;
            myPosition++;
            break;
        case '"':
        case '\'':
            //Start of quotation
            theStartPos = myPosition+1;
            theEndPos   = findEndString();
            myTooken->myType = TOKEN_QUOTEDTEXT;
            myTooken->myData = myData.substr(theStartPos, (theEndPos-theStartPos));
            myPosition = theEndPos+1;
            break;
        case '/':
            //End of tag tag?
            if(myData.at(myPosition+1) != '>')
            {
                std::cerr << "What? ("<< myData.at(myPosition+1) <<")" << std::endl;
                delete myTooken;
                throw;
            }
            myTooken->myType = TOKEN_EMPTYELEM;
            myPosition += 2;
            break;
        case '=':
            //Equal
            myTooken->myType = TOKEN_EQUAL;
            myPosition++;
            break;
        default: //Everything else...
            theStartPos = myPosition;
            theEndPos   = findEndString();
            myTooken->myType = TOKEN_TEXT;
            myTooken->myData = myData.substr(theStartPos, (theEndPos-theStartPos));
            myPosition = theEndPos;
    };

    return myTooken;
}
